/* console.c
 *  routines to write to the KIMS console screen
 *  with the basic 16x22 character set, stripped down
 *  to 12x22 (128x52 character screen)
 *  NOTE:  a 12x18 charset would be nicer... (128x64 screen)
 */




struct con_info {
	short	x;	/* current cursor x position */
	short	y;	/* current cursor y position */
	short	wid;	/* total width of window */
	short	hgt;	/* total height of window */
	short	rmin;	/* minimum (top) of scroll region */
	short	rmax;	/* maximum (bottom) of scroll region */
	short	attr;	/* 1:highlight 4:uline 5:blink 7:inverse */
	short   mode;	/* 16 mode bits */
	short	qmode;	/* 16 mode bits for ? modes */
	short	old_x;	/* old x position (cursor position) */
	short	old_y;	/* old y position (cursor position) */
	short	cursor_level;	/* nesting of cursor_on/off requests */ 
	char	charset;/* charset selector */
	char	g[2];	/* g0 and g1 charset selectors */
	char	fcolor;	/* foreground color */
	char	bcolor;	/* background color */
	short	*gbuf;	/* 1536 pixel graphics buffer */
	char 	**cbuf;	/* array of character buffers */
	int	*cbsize; /* size of each buffer */
};


#define FONTX	0
#define FONTY	1280
#define CHRWID 12
#define CHRHGT 22

#define ESCMAX 128

struct con_info con, con_save;
char escstr[ESCMAX];	/* string to store escape sequences */
static int escpos;	/* position in escstr */
short *screen;		/* array with attributes and characters */
short cursor[4] = {0xfffc, 0xfffc, 0xfffc, 0xfffc};

static int gr_wid;	/* graphics block width */
static int gr_hgt;	/* graphics block height */
static int gr_x;	/* graphics block x position */
static int gr_y;	/* graphics block y position */
static int gr_xofs;	/* graphics block x offset */
static int gr_yofs;	/* graphics block y offset */
static int gr_repeat;	/* pixel repeat count */
static int cb_num;	/* character buffer number */
static int cb_ofs;	/* offset in current char buf */
static int cb_size;	/* size of current character buf */

#define con_putchar(x,y,c)	 gcopy(((c)&0x7f)<<4,FONTY+((c)/128)*CHRHGT, \
 					CHRWID,CHRHGT, \
					(x)*CHRWID,(y)*CHRHGT, \
					CHRWID,0,0,CHRHGT) 
#define con_drawcursor(x,y)	 gcopy(0,FONTY+CHRHGT*14,CHRWID,3, \
					(x)*CHRWID,(y)*CHRHGT+CHRHGT-3, \
					CHRWID,0,0,3);



/* get a character from the ROM charset, and copy it into the buffer 
 * video PROMS are only 16-bits, so only every other 16-bit word is valid 
 */
get_romchar(buf,ch) 
	short *buf;
	int ch;
{
int base, i;
	base = (*(short*)(0x20140168 + ch*4))*2 + 0x20140168;
	for(i=0; i<22; i++)
		buf[i] = *(short*)(base+i*4);
}


con_init(bufsize)
	int bufsize;
{
	escpos = 0;
	con.x = 0;
	con.y = 0;
	con.wid = 128;
	con.hgt = 52;
	con.rmin = 0;
	con.rmax = con.hgt-1;
	con.attr = 0;
	con.mode = 0;
	con.qmode = 0x80;
	con.charset = 1;
	con.g[0] = 0;
	con.g[1] = 0;
	con.fcolor = 7;
	con.bcolor = 0;
	con.old_x = 0;
	con.old_y = 0;
	con.cursor_level = 0;
	con.gbuf = (short*)smem_alloc(1536*2);
	con.cbuf = (char**)smem_alloc(64*4);
	con.cbsize = (int*)smem_alloc(64*4);
	gr_wid = 0;		/* disable graphics block mode */
	cb_size = 0;		/* disable character buffer */
	screen = (short*)smem_alloc(2*con.wid*con.hgt);
	avss_init();
	load_font();
	con_redraw();
}


load_font()
{
int i,n;
short buf[22], tmp[22];

	/* normal video */
	for(i = 0; i < 128; i++) {
		get_romchar(buf, i);
		gput(buf, i * 16, FONTY, 16, CHRHGT);
		get_romchar(buf, i+128);
		gput(buf, i * 16, FONTY + CHRHGT, 16, CHRHGT);
	}

	/* highlighted video */
	for(i = 0; i < 128; i++) {
		get_romchar(tmp, i);
		for(n = 0; n < 22; n++)
			buf[n] = tmp[n] | (tmp[n] << 1);
		gput(buf, i * 16, FONTY + CHRHGT*2, 16, CHRHGT);
		get_romchar(tmp, i+128);
		for(n = 0; n < 22; n++)
			buf[n] = tmp[n] | (tmp[n] << 1);
		gput(buf, i * 16, FONTY + CHRHGT*3, 16, CHRHGT);
	}

	/* subscript */
	for(i = 0; i < 128; i++) {
		get_romchar(tmp, i);
		for(n = 0; n < 22; n++)
			buf[n] = tmp[(n+19) % 22];
		gput(buf, i * 16, FONTY + CHRHGT*4, 16, CHRHGT);
	}

	/* superscript */
	for(i = 0; i < 128; i++) {
		get_romchar(tmp, i);
		for(n = 0; n < 22; n++)
			buf[n] = tmp[(n+3) % 22];
		gput(buf, i * 16, FONTY + CHRHGT*6, 16, CHRHGT);
	}

	/* underscore */
	for(i = 0; i < 128; i++) {
		get_romchar(buf, i);
		for(n = 21; n < 22; n++)
			buf[n] = 0xffff;
		gput(buf, i * 16, FONTY + CHRHGT*8, 16, CHRHGT);
	}

	/* italic */
	for(i = 0; i < 128; i++) {
		get_romchar(tmp, i);
		for(n = 0; n < 22; n++)
			buf[n] = tmp[n] << ((21-n)/5);
		gput(buf, i*16, FONTY + CHRHGT*10, 16, CHRHGT);
	}

	/* inverse video */
	for(i = 0; i < 128; i++) {
		get_romchar(tmp, i);
		for(n = 0; n < 22; n++)
			buf[n] = (tmp[n] | (tmp[n] << 1)) ^ 0xffff;
		gput(buf, i*16, FONTY + CHRHGT*14, 16, CHRHGT);
	}
}

	

con_redraw()
{
int x,y;
int ch;
	con_cursor_off();
	for(x=0; x<con.wid; x++) for(y=0; y<con.hgt; y++) {
		ch = screen[x + y*con.wid];
		con_putchar(x, y, ch);
	}
	con_cursor_on();
}



/*****

con_scr_save(buf)
	short *buf;
{
int x,y;
	for(y=0; y<con.hgt; y++) for(x=0; x<con.wid; x++) 
		*buf++ = screen[x + y*con.wid];
}


con_scr_restore(buf)
	short *buf;
{
int x,y;
int ch;
	con_cursor_off();
	for(y=0; y<con.hgt; y++) for(x=0; x<con.wid; x++) {
		ch = screen[x + y*con.wid] = *buf++;
		con_putchar(x, y, ch);
	}
	con_cursor_on();
}	

******/


rawcon_clear(x1,y1,x2,y2)
	int x1,y1,x2,y2;
{
	con_cursor_off();
	gclear(x1*CHRWID,y1*CHRHGT,	
		(x2-x1+1)*CHRWID,(y2-y1+1)*CHRHGT);
	con_cursor_on();
}


con_clear(x1,y1,x2,y2)
	int x1,y1,x2,y2;
{
int x,y,d;
int ch;
	rawcon_clear(x1,y1,x2,y2);
	for(y=y1; y<=y2; y++) {
		d = y*con.wid + x1; 
		for(x=x1; x<=x2; x++)
			screen[d++] = 0;
	}
}


con_cls()
{
	con_cursor_off();
	con_clear(0,0,con.wid-1,con.hgt-1);
	con_gotoxy(0, 0);
	con_cursor_on();
}


int con_getheight()
{
	return(con.hgt);
}


int con_getwidth()
{
	return(con.wid);
}


void con_setheight(hgt)
	int hgt;
{
	con.hgt = hgt;
}

int con_setwidth(wid)
	int wid;
{
	con.wid = wid;
}


con_scroll_up(x1,y1,x2,y2,n)
	int x1,y1,x2,y2,n;
{
int x,y;
int oldch;
int s,d;
	con_cursor_off();
	if (n)
		scroll_up(x1*CHRWID,y1*CHRHGT,(x2+1)*CHRWID-1,
			(y2+1)*CHRHGT-1,CHRHGT*n);
	for(y=y1; y<=y2-n; y++) {
		s = (y+n)*con.wid + x1;
		d = y*con.wid + x1;
		for(x=x1; x<=x2; x++)
			screen[d++] = screen[s++];
	}
	for(y=y2-n+1; y<=y2; y++) {
		d = y*con.wid + x1; 
		for(x=x1; x<=x2; x++)
			screen[d++] = 0;
	}
	con_gotoxy(con.x,con.y);
	con_cursor_on();
}


con_scroll_down(x1,y1,x2,y2,n)
	int x1,y1,x2,y2,n;
{
int x,y;
int oldch;
int s,d;
	con_cursor_off();
	for(y=y2; y>=y1+n; y--) {
		s = (y-n)*con.wid + x1;
		d = y*con.wid + x1;
		for(x=x1; x<=x2; x++)
			screen[d++] = screen[s++];
	}
	for(y=y1; y<y1+n; y++) {
		d = y*con.wid + x1; 
		for(x=x1; x<=x2; x++)
			screen[d++] = 0;
	}
	if (n)
		scroll_down(x1*CHRWID,y1*CHRHGT,(x2+1)*CHRWID-1,
			(y2+1)*CHRHGT,CHRHGT*n);
	con_gotoxy(con.x,con.y);
	con_cursor_on();
}


con_scroll_left(x1,y1,x2,y2,n)
	int x1,y1,x2,y2,n;
{
int x,y;
int oldch;
int s,d;
	con_cursor_off();
	for(y=y2; y>=y1; y--) {
		s = y*con.wid;
		for(x=x1; x<=x2-n; x++) {
			oldch = screen[s+x] = screen[s+x+n];
			con_putchar(x,y,oldch);
		}
		screen[s+x2] = 0;
		con_putchar(x2,y,0);
	}
	con_gotoxy(con.x,con.y);
	con_cursor_on();
}


con_scroll_right(x1,y1,x2,y2,n)
	int x1,y1,x2,y2,n;
{
int x,y;
int oldch;
int s,d;
	con_cursor_off();
	for(y=y2; y>=y1; y--) {
		s = y*con.wid;
		for(x=x2; x>=x1+n; x--) {
			oldch = screen[s+x] = screen[s+x-n];
			con_putchar(x,y,oldch);
		}
		screen[s+x1] = 0;
		con_putchar(x1,y,0);
	}
	con_gotoxy(con.x,con.y);
	con_cursor_on();
}


con_puts(str)
	char *str;
{
	con_write(str, strlen(str));
}


con_printf(fmt, str)
	char *fmt, *str;
{
char buf[512];
	vsprintf(buf, fmt, &str);
	con_puts(buf);
}


con_putc(ch)
	char ch;
{
	con_write(&ch, 1);
}


con_gotoxy(x, y)
	int x, y;
{
int oldch;
	con_cursor_off();
	con.x = x;
	con.y = y;
	if (con.x < 0)
		con.x = 0;
	if (con.x >= con.wid) {
		con.x = 0;		
		con.y++;
	}
	if (con.y < con.rmin) {
		con.y = con.rmin;
		con_scroll_down(0,con.rmin,con.wid-1,con.rmax,1);
	}
	if (con.y > con.rmax) {
		con.y = con.rmax;
		con_scroll_up(0,con.rmin,con.wid-1,con.rmax,1);
	}
	con_cursor_on();
}


con_cursor_off()
{
int oldch;
	if (!con.cursor_level++) {
		oldch = screen[con.old_x + con.old_y * con.wid];
		con_putchar(con.old_x, con.old_y, oldch);
	}
}


con_cursor_on()
{
	if (!(--con.cursor_level)) {
		con.old_x = con.x;
		con.old_y = con.y;
		con_drawcursor(con.old_x, con.old_y);
	}
}



rawcon_write(x, y, attr, buf, len)
	int x, y, attr;
	unsigned char *buf;
	int len;
{
int yp;
	yp = FONTY + attr*CHRHGT*2;
	while(len--) {
		gcopy(((*buf)&0x7f) << 4, yp+((*buf)/128)*CHRHGT,
			CHRWID, CHRHGT,
			x*CHRWID, y*CHRHGT,
			CHRWID, 0, 0, CHRHGT);
		buf++;
		x++;
	}
}



con_write(buf, len)
	char *buf;
	int len;
{
int ch;
int oldch;
char *cp;
int i, n;
	con_cursor_off();
	while(len--) {	
		ch = *buf++;
		if (cb_size) {
			con.cbuf[cb_num][cb_ofs++] = ch;
			if(cb_ofs >= cb_size) 
				cb_size = 0;
		} else if (gr_wid && ch != 27)
			graph_block(ch);
		else if (escpos) {
			if (escpos >= ESCMAX) {
				escpos = 0;	/* in case of overflow */
				break;		/* just get out */
			}
			escstr[escpos++] = ch;
			if (escpos == 2 && (ch == '#' || ch == '('
			 || ch == ')'))
				continue;
			if (escpos == 2 && ch != '[') {
				esc_seq_ncsi(escstr,escpos);
				escpos = 0;
				continue;
			}
			if (escpos == 3 && (escstr[1] == '#' 
		          || escstr[1] == '(' || escstr[1] == ')')) {
				esc_seq_ncsi(escstr,escpos);
				escpos = 0;
				continue;
			}
			if(((ch < '0' || ch > '9') && ch != ';' 
			   && ch != '[' && ch != '?')) {
				esc_seq_csi(escstr,escpos);
				escpos = 0;
			}
		} else if (ch >= 0x20 && ch <= 0x7e) {
			if(con.mode & 0x10)	/* insert mode */
				con_scroll_right(con.x,con.y,
					con.wid-1,con.y,1);
			if (con.g[con.charset] && (ch > 95))
				ch -= 95;
			con_putchar(con.x, con.y, 
				screen[con.x + con.y * con.wid] = 
					(con.attr << 8) | ch);
			if (++con.x >= con.wid && !(con.qmode & 0x80))
				con.x = con.wid-1;
			con_gotoxy(con.x, con.y);
		} else if (ch == 8) {
			con_gotoxy(--con.x, con.y);
		} else if (ch == 9) {
			if(con.mode & 0x10)	/* insert mode */
				con_scroll_right(con.x,con.y,
					con.wid-1,con.y,
					(con.x & ~7)+8-con.x);
			con.x = (con.x & ~7)+8;
			con_gotoxy(con.x, con.y);
		} else if (ch == 27) {
			escstr[escpos++] = ch;
		} else if (ch >= 10 && ch <= 12) {
			if (con.y == con.rmax) { /* LF search */
				cp = buf;
				n = 1;
				i = len;
				while(i-- && *cp && *cp != 27 
				&& n < con.rmax - con.rmin - 4) {
					if (*cp >= 10 && *cp <= 12)
						n++;
					cp++;
				}
				con_scroll_up(0,con.rmin,con.wid-1,con.rmax,n);
				con.y += (1 - n);
			} else 
				con_gotoxy(con.x, ++con.y);
		} else if (ch == 13) {
			con_gotoxy(0, con.y);
		} else if (ch == 7) {
			/* rawcon_write(116,0,5,"** beep! **", 11);  */
			kb_write("\247",1);  
		} else if (ch == 14) {
			con.charset = 1;
		} else if (ch == 15) {
			con.charset = 0;
		} else
			ctrl_unknown(ch);
	}
	con_cursor_on();
}


static int esc_getarg(def)
	int def;
{
int n;
char ch;
int ok;
int base;
int extend;
	extend = 0;
	base = 10;
	n = 0;
	ok = 0;
	ch = escstr[escpos];
/*
	if (ch == '(') {
		base = 16;
		ch = escstr[++escpos];
	}	
*/
	if(ch=='?') {
		extend = 256;
		ch = escstr[++escpos];
	}
	while ((ch >= '0' && ch <= '9') 
		|| (base == 16 && ((ch >= 'a' && ch <= 'f') 
			|| (ch >= 'A' && ch <= 'F')) )) {
		if (ch >= 'a' && ch <= 'f')
			ch = (ch - 'a' + 10 + '0');
		else if (ch >= 'A' && ch <= 'F')
			ch = (ch - 'A' + 10 + '0');
		n = n * 10 + (ch - '0');
		ch = escstr[++escpos];
		ok++;
	}
/*
	if (ch == ')')
		ch = escstr[++escpos];
*/
	if (!ok)
		if (ch == ';' || def >= 0)
			n = def;
		else
			n = -1;
	if (ch == ';')
		escpos++;
	return(n+extend);
}



esc_seq_csi(str,len)
	char *str;
	int len;
{
int x,y,a,b,c;
	escstr[escpos] = '\0';
	escpos = 2;
	switch(str[len-1]) {
		case 'H': case 'f':
			con.y = esc_getarg(1) - 1;
			con.x = esc_getarg(1) - 1;
			break;
		case 'J':
			a = esc_getarg(0);
			switch(a) {
			    case 0: 
				con_clear(con.x,con.y,con.wid-1,con.y);
				con_clear(0,con.y+1,con.wid-1,con.hgt-1);
				break;
			    case 1:
				con_clear(0,0,con.wid-1,con.y-1);
				con_clear(0,con.y,con.x,con.y);
				break;
			    case 2:
				con_clear(0,0,con.wid-1,con.hgt-1);
				break;
			}
			break;
		case 'K':
			a = esc_getarg(0);
			switch(a) {
			    case 0: 
				con_clear(con.x,con.y,con.wid-1,con.y);
				break;
			    case 1:
				con_clear(0,con.y,con.x,con.y);
				break;
			    case 2:
				con_clear(0,con.y,con.wid-1,con.y);
				break;
			}
			break;
		case 'L':
			a = esc_getarg(1);
			con_scroll_down(0,con.y,con.wid-1,con.rmax,a);
			break;
		case 'M':
			a = esc_getarg(1);
			con_scroll_up(0,con.y,con.wid-1,con.rmax,a);
			break;
		case '@':
			a = esc_getarg(1);
			con_scroll_right(con.x,con.y,con.wid-1,con.y,a);
			break;
		case 'P':
			a = esc_getarg(1);
			con_scroll_left(con.x,con.y,con.wid-1,con.y,a);
			break;
		case 'm':
			x = 0;
			while ((a = esc_getarg(-1)) >= 0) {
				x++;
				if (a == 0)
					con.attr = 0;
				else if (a <= 15)
					con.attr  =   a;
				else if (a >= 20 && a <= 27)
					con.attr = 0;
				else if (a >= 30 && a <= 37)
					con.fcolor = a;
				else if (a >= 40 && a <= 47)
					con.bcolor = a;
				else if (a >= 50 && a <= 57)
					con.fcolor = a + 8;
				else if (a >= 60 && a <= 67)
					con.bcolor = a + 8;
			}
			if (x == 0)
				con.attr = 0;
			break;
		case 's':
			con_save = con;
			con_gotoxy(con.x, con.y);
			break;
		case 'u':
			con = con_save;
			con_gotoxy(con.x, con.y);
			break;
		case 'r':
			a = esc_getarg(1);
			b = esc_getarg(con.hgt);
			con.rmin = a-1;
			con.rmax = b-1;
			break;
		case 'A':
			a = esc_getarg(1);
			con_gotoxy(con.x, con.y-a);
			break;
		case 'B':
			a = esc_getarg(1);
			con_gotoxy(con.x, con.y+a);
			break;
		case 'C':
			a = esc_getarg(1);
			con_gotoxy(con.x+a, con.y);
			break;
		case 'D':
			a = esc_getarg(1);
			con_gotoxy(con.x-a, con.y);
			break;
		case 'h':
			while ((a = esc_getarg(-1)) >= 0) {
				if (a < 256)
					con.mode |= (1 << a);
				else
					con.qmode |= (1 << (a&0xff));
			} 
			break;
		case 'l':
			while ((a = esc_getarg(-1)) >= 0) {
				if (a < 256)
					con.mode &= (0xffff ^ (1 << a));
				else
					con.mode &= (0xffff ^ (1 << (a&0xff)));
			}
			break;
		case 'g':
			gr_x = esc_getarg(0);
			gr_y = esc_getarg(0);
			gr_wid = esc_getarg(0);
			gr_hgt = esc_getarg(0);
			gr_xofs = 0;
			gr_yofs = 0;
			gr_repeat = 0;
			break;
		case 'b':
			cb_num = esc_getarg(0);
			if(cb_num > 256) {
				cb_num &= 0xff;
				cb_size = esc_getarg(0);
				cb_ofs = 0;
				con.cbuf[cb_num] =
					(char*)smem_alloc(cb_size);
				con.cbsize[cb_num] = cb_size;
			} else 
				con_write(con.cbuf[cb_num], 
					con.cbsize[cb_num]);
			break;
		default:
			escpos = 0;
			esc_unknown(escstr);
	}	
}


esc_seq_ncsi(str,len)
	char *str;
	int len;
{
int x,y,a,b,c;
	escstr[escpos] = '\0';
	escpos = 0;
	switch(str[1]) {
		case 'D':
			con_gotoxy(con.x, con.y+1);
			break;
		case 'E':
			con_gotoxy(0, con.y+1);
			break;
		case 'H':
			break;
		case 'M':
			con_gotoxy(con.x, con.y-1);
			break;
		case '7':
			con_save.x = con.x;
			con_save.y = con.y;
			con_save.attr = con.attr;
			con_save.fcolor = con.fcolor;
			con_save.bcolor = con.bcolor;
			con_gotoxy(con.x, con.y);
			break;
		case '8':
			con.x = con_save.x;
			con.y = con_save.y;
			con.attr = con_save.attr;
			con.fcolor = con_save.fcolor;
			con.bcolor = con_save.bcolor;
			con_gotoxy(con.x, con.y);
			break;
		case '=':
			break;
		case '>':
			break;
		case '#':
			break;
		case '(':
			con.g[0] = (str[2] == '0' || str[2] == '2');
			break;
		case ')':
			con.g[1] = (str[2] == '0' || str[2] == '2');
			break;
		default:
			escpos = 0;
			esc_unknown(escstr);
	}
}


print_esc(str)
	char *str;
{
char ch;
char *cp;
char buf[10];
	cp = str;
	while(ch = *cp++) {
		if(ch < 32) {
			con_write("^", 1);
			ch += 64;
		}
		con_write(&ch, 1);
	}
/* hex dump disabled, to remove stdio library (sprintf)
	cp = str;
	con_write("(", 1);
	while(ch = *cp++) {
		sprintf(buf, "%2x", ch);
		con_write(buf, 2);
		if (*cp)
			con_write(" ", 1);
	}
	con_write(")", 1);
*/
}


graph_block(ch)
	unsigned char ch;
{
int data;
	data = ch - 0x20;
	if (data < 0)
		return;
	if (gr_repeat == -1) {
		gr_repeat = data;
		return;
	}
	if (data == 64)
		data = 0;
	if (data == 65)
		gr_repeat = -1;
	else {
		if(gr_repeat < 1)
			gr_repeat = 1;
		while(gr_repeat--) {
			con.gbuf[gr_xofs++] = ((data >> 4) & 3);
			con.gbuf[gr_xofs++] = ((data >> 2) & 3);
			con.gbuf[gr_xofs++] = (data & 3);
			if (gr_xofs >= gr_wid) {
				gput_z(con.gbuf, gr_x, gr_y+gr_yofs, 
					gr_xofs, 1);
				gr_xofs = 0;
				gr_yofs++;
			}
			if (gr_yofs >= gr_hgt) {
				gr_wid = 0;
				return;
			}
		}
		gr_repeat = 0;
	}
}
