/*  jmon.c
 *   object-oriented monitor program  
 *   Project JakeDev
 *   started December 10th, 1990
 *   by Frank Barrus and members of Computer Science House
 */

/* monitor commands:
 *  <adr>,<len>?<type>  display range of data
 *         <type> = b     byte
 *                  c     character (ascii)
 *                  d     decimal
 *                  x     hexidecimal
 *                  i     instructions
 *  <adr>$r             run at specified address 
 *  $q                  quit (return to terminal mode)
 *  &H                  halt system 
 *  &X			download straight binary
 *  &D			download with packets
 */


#include "types.h"


int num_of_str(str)
	char *str;
{
	return(atoi(str));
}



#define printable(ch)    (ch>126||ch<32?'~':ch)



uint32 printval(buf, adr, dsize, dtype)
char *buf;
char *adr;
int  dsize;
char dtype;
{
uint32 i;
int n10,nc;
   switch(dsize) {
      case 1:  i = *((uint8*)adr++); n10=3; break;
      case 2:  i = *((uint16*)adr++); n10=5; break;
      case 4:  i = *((uint32*)adr++); n10=10; break;
   }
   switch(dtype) {
      case 'a': nc = 1; *buf++ = printable(i); break;
      case 'd': nc = sprintf(buf,"%d", i, n10); break;
      case 'x': nc = sprintf(buf,"%x", i, dsize*2); break;
      case 'b': nc = sprintf(buf,"%b", i, dsize*8); break;
      case 'o': nc = sprintf(buf,"%o", i, dsize*8/3); break;
   }
   return(nc); 
}


sprintn()
{
return;
}

uint32 hexdump(adr, size, dsize, dtype)
uint32 adr;
uint32 size;
int  dsize;
char dtype;
{
  uint32  i;
  int     x;
  int     s;
  char buf[80];

  i = adr;
  while(i < adr + size) {
    x = sprintn(buf, i, 16, 8);
    buf[x++] = ':';
    buf[x++] = ' ';
    s = 0;
    while(x+s < 78 && i < adr + size) {
       s = printval(&buf[x], i, dsize, dtype);
       x += s;
       if (dtype != 'a')
          buf[x++] = ' ';
       i += dsize;
    } 
    buf[x++] = '\n';
    buf[x++] = '\0';
    puts(buf);
  }
  return(i); 
}



jmon()
{
char str[80];
char buf[80];
uint32 p1,p2;
int n;
int i;
char cmdtype;
char cmd;
int dsize;
char dtype;
boolean quit = FALSE;
   puts("\nJMON v0.1\n");
   p1 = (uint32)buf;
   p2 = 64;
   cmdtype = '?';
   cmd = 'x';
   dsize = 4;
   while(!quit) {
      sprintn(buf,p1,16,8);
      buf[8] = ',';
      sprintn(&buf[9],p2,16,4);
      buf[13] = cmdtype;
      buf[14] = cmd; 
      buf[15] = dsize + '0';
      buf[16] = '>';
      buf[17] = '\0';
      puts(buf);
      n = creadline(str, 32, '\n');
      if (n>0) {
         if(str[0] == ',') {
            i=get_word(buf,str,1);
            p2=num_of_str(buf);
         } else {
            i=get_word(buf,str,0);
            if(buf[0]) {
               p1=num_of_str(buf) & ~1;
               if (str[i] == ',') {
                  i=get_word(buf,str,i+1);
                  if(i>0) {
                     p2=num_of_str(buf);
                  }
               }
            }
         }
         if (str[i]) 
            cmdtype = str[i++];
         if (str[i])
            cmd = str[i++];
      }
      switch(cmdtype) { 
        case '?':
           switch(cmd) {
             case 'a':
             case 'x': 
             case 'd': 
             case 'b':
	     case 'o': dtype = cmd; break;
             case 'w': 
             case 'W':
                do {
                   i = get_word(buf,str,i);
                   if (buf[0]) {
		     n = num_of_str(buf);
		     if (cmd == 'w')
                        *(uint16*)p1 = n;
                     else
                        *(uint32*)p1 = n; 
                     p1 += 2*(1+(cmd=='W'));
                   }
                } while(buf[0]); break; 
             default:  puts("bad modifier: ");
                       cputc(cmd);
                       cputc('\n');
           } 
           if (cmd != 'w' && cmd != 'W')
             p1 = hexdump(p1,p2,dsize,dtype); break;
        case '$': 
           switch(cmd) {
             case '1':
             case '2':
             case '4': dsize = cmd - '0'; break;
             case 'q': quit = TRUE; break;
           }
           break;
        case '&':
           switch(cmd) {
/*
             case 'H': shutdown(); break;
             case 'd': cdread(p1,p2); break;
             case 'D': cread(p1,p2); break;
             case 'u': cdwrite(p1,p2); break;
             case 'U': cwrite(p1,p2); break;
*/
             case 'g': (*(fptr*)(p1))(); break;
           }
           cmd = ' ';
           break;
        case ' ': break;
        default: puts("bad syntax: ");
                 cputc(cmdtype);
                 cputc('\n');
      }
   }
   puts("*** mx\n");
}

