/* kims_serial.s
 *   various serial I/O routines
 *   also does keyboard scan code to ascii conversions
 */


	.set inbuf_size,16



/* serial read into buffer  
 *   serial_read(port,buf,size,delay)
 */
.globl _serial_read
_serial_read:	.word 0x7e
		movl	4(ap),r2	/* channel number */
		movl	8(ap),r5	/* buffer address */
		movl	12(ap),r3	/* number of bytes */
		clrl	r4		/* count of characters read */
sreadloop:	movl	16(ap),r6	/* delay count */
		ashl	$10,r6,r6
sdelayloop:	calls	$0, _serial_poll
		movl	in_buf_rpos[r2],r1
		cmpl	r1,in_buf_wpos[r2]
		bneq	sgotit
		sobgtr	r6,sdelayloop
		brb	sreaddone
sgotit:		mull3	r2, $inbuf_size, r0
		addl2	r1, r0
		movb	in_buf(r0),(r5)+	/* store it */
		incl	r1
		cmpl	r1, $inbuf_size
		blss	snowrap
		clrl	r1
snowrap:	movl	r1, in_buf_rpos[r2]
		incl	r4		/* increase character count */
		sobgtr	r3,sreadloop	/* and get some more */
sreaddone:	movl	r4,r0		/* return num chars read */
		ret


/* poll serial ports for incoming data 
 *  - should be called from interrupt routine 
 */
.globl _serial_poll
_serial_poll:	.word 0x0e
		movl	$0x200a0000,r2
hpoll_loop:	movzwl	4(r2),r1
		bbc	$15,r1,hpoll_done
		extzv	$8,$2,r1,r0
		bitw	$0x6000,r1
		bneq	hpoll_loop
		movzbl	r1,r1
		tstb	r0
		bneq	notkb
		bsbb	kbd_conv
		cmpl	r1,$-1
		beql	hpoll_loop
notkb:		mull3	r0, $inbuf_size, r3
		addl2	in_buf_wpos[r0], r3
		movb	r1,in_buf(r3)
		incl	in_buf_wpos[r0]
		cmpl	in_buf_wpos[r0], $inbuf_size
		blss	hpoll_loop
		clrl	in_buf_wpos[r0]
		brb	hpoll_loop
hpoll_done:	ret


kbd_conv:
		cmpb	r1,$0xaf
		bneq	kbd1
		movb	$1,shift_mode
		movl	$-1,r1
		rsb
kbd1:		cmpb	r1,$0xae
		bneq	kbd2
		movb	$2,shift_mode
		movl	$-1,r1
		rsb
kbd2:		cmpb 	r1,$0xb0
		bneq	kbd3
		movb	$3,shift_mode
		movl	$-1,r1
		rsb
kbd3:		cmpb	r1,$0xb3
		bneq	kbd4
		movb	$0,shift_mode	
		movl	$-1,r1
		rsb	
kbd4:		movl	*$0x20040084, r3
		subb2	$0xbc, r1
		movzbl	(r3)[r1],r1
		movl	*$0x20040080, r3
		ashl	$2,r1,r1
		addb2	shift_mode,r1
		movzbl	(r3)[r1],r1
		rsb



shift_mode:	.byte	0
		
in_buf:		.org	in_buf+inbuf_size
	i1:	.org	i1+inbuf_size
	i2:	.org	i2+inbuf_size
	i3:	.org	i3+inbuf_size
in_buf_rpos:	.long	0,0,0,0
in_buf_wpos:	.long	0,0,0,0



/* serial write from buffer
 *   serial_write(port,buf,size)
 */
.globl _serial_write
_serial_write:	.word 0x7e
		movl	4(ap),r2	/* channel number */
		movl	8(ap),r5	/* buffer address */
		movl	12(ap),r3	/* number of bytes */
		clrl	r4		/* count of characters written */
		movl	$0x200a0000,r6
		clrw	8(r6)
		ashl	r2,$1,r0
		bisw3	$0x0100,r0,8(r6)/* select line */
stready:	bbc	$15,(r6),stready	/* trans ready? */
swriteloop:	movb	(r5)+,r0
		movzbw	r0,0x0c(r6)	/* transmit it */
swready:	bbc	$15,(r6),swready	/* ready? */
		incl	r4		/* increase character count */
		sobgtr	r3,swriteloop	/* and get some more */
		movl	r4,r0		/* return num chars read */
		ret
