/*  parse.c
 *   utilities for parsing input 
 *   author: Frank Barrus
 */


#include "types.h"

#define DEF_NUMBASE 16
#define digit_value(ch)  (ch<='9'?ch-'0':10+(ch>='A'&&ch<='Z'?ch-'A':ch-'a'))
#define printable(ch)    (ch>126||ch<32?'~':ch)

#define  is_num(x)    ((x >= '0' && x <= '9') || x == '-')
#define  is_word(x)   ((x >= 'A' && x <= 'Z') || (x >= 'a' && x <= 'z'))
#define  is_sym(x)    (is_num(x) || is_word(x))
 

/* numbers are accepted in the following format:
 *   ddddd   default base
 *   0dddd   octal
 *   0xddd   hex
 *   0tddd   decimal
 *   0nddd   binary
 */

uint32  num_of_str(str)
 char *str;
{
int base, x;
uint32 num = 0;
boolean neg = FALSE;
  base = DEF_NUMBASE;
  if(*str == '-') {
    neg = TRUE;
    str++;
  }
  if(*str == '0') {
    base = 8;
    str++;
    switch(*str) {
      case 'x': base = 16; str++; break;
      case 't': base = 10; str++; break;
      case 'n': base = 2; str++; break;
    }
  }
  while((x = digit_value(*str)) >= 0 && (x < base)) {
    num = (num*base)+x;
    str++;
  }
  if (neg)
    return(-num);
  else 
    return(num);
}


int get_word(wbuf, str, pos)
char *wbuf;
char *str;
int  pos;
{
int wpos = 0;
   while(str[pos] == ' ')
      pos++;
   while(is_sym(str[pos]))
      wbuf[wpos++] = str[pos++];
   wbuf[wpos] = '\0';
   return(pos);
}

