/*  avss.c
 *   actor video system routines
 */


#include "types.h"
#include "kims_avss.h"


void avss_id_write(adr,value)
	int adr;
	int value;
{
	while(!(mem16(ADDER+STAT) & bit(4)))
		continue;
	while(!(mem16(ADDER+STAT) & bit(6)))
		continue;
	mem16(ADDER+IDD) = value;
	mem16(ADDER+CMD) = 0x100 + adr;
}


uint16 avss_init_tab[50] = {
	0x60, 0x01,	/* select plane 0 */
	0x83, 0x01,	/* address 1 */
	0x60, 0x02,	/* select plane 1 */
	0x83, 0x00,	/* address 0 */
	0x60, 0x03,	/* planes 0 and 1 */
	0x84, 0xca,	/* f(0) = s */
	0x85, 0xce,	/* f(1) = not(s) */
	0x86, 0xce,	/* f(2) = d or s */
	0x87, 0xc6,	/* f(3) = d xor s */
	0x88, 0xffff,	/* mask 1 & 2 */
	0x80, 0x00,	/* one plane res */
	0x81, 0x0c,	/* 16-bit bus */
	0x8a, 0xffff,	/* source register */
	0x8e, 0x0000,	/* background */
	0x8f, 0xffff,	/* foreground */
	0x90, 0x29,	/* CSR0 */
	0x60, 0x01,	/* select plane 0 */
	0x94, 0x35,	/* CSR4 - SRC1 */
	0x60, 0x02,	/* select plane 1 */
	0x94, 0x21,	/* CSR4 - SRC1 */
	0x60, 0x03,	/* select both planes */
	0x91, 0x05,	/* CSR1 - SRC2 */
	0x92, 0x00,	/* CSR2 - DEST1 */
	0x95, 0x09,	/* CSR5 */
	0x96, 0x00	/* CSR6 */
};	


void avss_init()
{
int i, t;
	t = 30000;
	while(!(mem16(ADDER+STAT) & bit(3)) && t--)
		continue;
	for(i=0; i<50; i += 2)
		avss_id_write(avss_init_tab[i], avss_init_tab[i+1]);

}




void avss_cursor_move(x,y)
	int x,y;
{
	mem16(CUR_XPOS) = (x/2) + 8;
	mem16(CUR_YPOS) = y + 40;
	mem16(VID_CSR) = (mem16(VID_CSR) & ~bit(5)) | ((x&1) << 5);
}

void avss_cursor_mode(mode)
	int mode;
{
	switch(mode) {
		case 0:
			mem16(CUR_CMD) = 0xe000;
			break;
		case 1:
			mem16(CUR_CMD) = 0xe011;
			break;
		case 2:
			mem16(CUR_CMD) = 0xe014;
			break;
		case 3:
			mem16(CUR_CMD) = 0xe015;
			break;
	}
}



void gline(x1, y1, x2, y2, func, source)
	int x1, y1, x2, y2, func, source;
{
int t;
int dx,dy;
	t = 30000;
	while(!(mem16(ADDER+STAT) & bit(3)) && t--)
		continue;
	avss_id_write(0x88, 0xffff);	/* mask 1 & mask 2 */
	avss_id_write(0x8a, source); 	/* source register */
	avss_id_write(0x8e, 0x0000);	/* background */
	avss_id_write(0x8f, 0xffff);	/* foreground */
	while(!(mem16(ADDER+STAT) & bit(4)))
		continue;
	while(!(mem16(ADDER+STAT) & bit(6)))
		continue;
	dx = (x2-x1);
	dy = (y2-y1);
	if (dx < 0)
		dx--;
	else
		dx++;
	if (dy < 0)
		dy--;
	else
		dy++;
	mem16(ADDER+DXO) = x1;
	mem16(ADDER+DYO) = y1;
	mem16(ADDER+FDX) = dx;
	mem16(ADDER+FDY) = dy;
	mem16(ADDER+SDX) = 1;
	mem16(ADDER+SDY) = 1;
	mem16(ADDER+FSC) = 0x1fff;
	mem16(ADDER+SSC) = 0x1fff;
	mem16(ADDER+MDE) = 0xc0;
	mem16(ADDER+CMD) = 0x6c4 + (func << 4);
}



void gclear(dx, dy, dfvx, dsvy)
	int dx, dy, dfvx, dsvy;
{
int t;
	t = 30000;
	while(!(mem16(ADDER+STAT) & bit(3)) && t--)
		continue;
	avss_id_write(0x88, 0xffff);	/* mask 1 & mask 2 */
	avss_id_write(0x8a, 0x0000); 	/* source register */
	avss_id_write(0x8e, 0x0000);	/* background */
	avss_id_write(0x8f, 0xffff);	/* foreground */
	while(!(mem16(ADDER+STAT) & bit(4)))
		continue;
	while(!(mem16(ADDER+STAT) & bit(6)))
		continue;
	mem16(ADDER+DXO) = dx;
	mem16(ADDER+DYO) = dy;
	mem16(ADDER+FDX) = dfvx;
	mem16(ADDER+FDY) = 0;
	mem16(ADDER+SDX) = 0;
	mem16(ADDER+SDY) = dsvy;
	mem16(ADDER+FSC) = 0x1fff;
	mem16(ADDER+SSC) = 0x1fff;
	mem16(ADDER+MDE) = 0xc0;
	mem16(ADDER+CMD) = 0x6c4;
}



void gcopy(sx, sy, sfvx, ssvy, dx, dy, dfvx, dfvy, dsvx, dsvy)
	int sx, sy, sfvx, ssvy, dx, dy, dfvx, dfvy, dsvx, dsvy;
{
int t;
	t = 30000;
	while(!(mem16(ADDER+STAT) & bit(3)) && t--)
		continue;
	avss_id_write(0x88, 0xffff);	/* mask 1 & mask 2 */
	avss_id_write(0x8a, 0xffff); 	/* source register */
	avss_id_write(0x8e, 0x0000);	/* background */
	avss_id_write(0x8f, 0xffff);	/* foreground */
	while(!(mem16(ADDER+STAT) & bit(4)))
		continue;
	while(!(mem16(ADDER+STAT) & bit(6)))
		continue;
	mem16(ADDER+SXO) = sx;
	mem16(ADDER+SYO) = sy;
	mem16(ADDER+FSDX) = sfvx;
	mem16(ADDER+SSDY) = ssvy;
	mem16(ADDER+DXO) = dx;
	mem16(ADDER+DYO) = dy;
	mem16(ADDER+FDX) = dfvx;
	mem16(ADDER+FDY) = dfvy;
	mem16(ADDER+SDX) = dsvx;
	mem16(ADDER+SDY) = dsvy;
	mem16(ADDER+FSC) = 0x1fff;
	mem16(ADDER+SSC) = 0x1fff;
	mem16(ADDER+MDE) = 0xc0;
	mem16(ADDER+CMD) = 0xec4;
}



gput(buf, x,y,wid,hgt)
	short *buf;
	int x,y,wid,hgt;
{
int t;
register int r11,r10,r9,r8;
	t = 30000;
	while(!(mem16(ADDER+STAT) & bit(3)) && t--)
		continue;
	avss_id_write(0x88, 0xffff);	/* mask 1 & mask 2 */
	avss_id_write(0x8a, 0xffff); 	/* source register */
	avss_id_write(0x8e, 0x0000);	/* background */
	avss_id_write(0x8f, 0xffff);	/* foreground */
	while(!(mem16(ADDER+STAT) & bit(4)))
		continue;
	while(!(mem16(ADDER+STAT) & bit(6)))
		continue;
	mem16(ADDER+SXO) = 0;
	mem16(ADDER+SYO) = 0;
	mem16(ADDER+FSDX) = wid;
	mem16(ADDER+SSDY) = hgt;
	mem16(ADDER+DXO) = x;
	mem16(ADDER+DYO) = y;
	mem16(ADDER+FDX) = wid;
	mem16(ADDER+FDY) = 0;
	mem16(ADDER+SDX) = 0;
	mem16(ADDER+SDY) = hgt;
	mem16(ADDER+FSC) = 0x1fff;
	mem16(ADDER+SSC) = 0x1fff;
	mem16(ADDER+MDE) = 0xc0;
	mem16(ADDER+CMD) = 0x7c1;
	r11 = ADDER+STAT;
	r10 = ADDER+IDD;
	r8 = (int)buf;
	r9 = r8 + ((wid*hgt)/16 + 1) * 2;
	asm("gpl1: bbc $6,(r11),gpl1");
	asm("gpl2: movw (r8)+,(r10)");
	asm("	   bbs $6,(r11),gpl2");	
	asm("      cmpl r8,r9");
	asm("      blss gpl1");
/*
	asm("gpl1: bbc $6,(r11),gpl1");
	asm("      movw (r8)+,(r10)");
	asm("      sobgtr r9,gpl1");
*/
/*
	while(i--) {
		while(!(mem16(ADDER+STAT) & bit(6)))
			continue;
		mem16(ADDER+IDD) = *buf++;
	}
*/

}


gput_z(buf, x,y,wid,hgt)
	short *buf;
	int x,y,wid,hgt;
{
register int r11,r10,r9,r8;
int t;
/*	t = 30000;
	while(!(mem16(ADDER+STAT) & bit(3)) && t--)
		continue;  
*/
	avss_id_write(0x88, 0xffff);	/* mask 1 & mask 2 */
	avss_id_write(0x8a, 0xffff); 	/* source register */
	avss_id_write(0x8e, 0x0000);	/* background */
	avss_id_write(0x8f, 0xffff);	/* foreground */
	while(!(mem16(ADDER+STAT) & bit(4)))
		continue;
	while(!(mem16(ADDER+STAT) & bit(6)))
		continue;
	mem16(ADDER+SXO) = 0;
	mem16(ADDER+SYO) = 0;
	mem16(ADDER+FSDX) = wid;
	mem16(ADDER+SSDY) = hgt;
	mem16(ADDER+DXO) = x;
	mem16(ADDER+DYO) = y;
	mem16(ADDER+FDX) = wid;
	mem16(ADDER+FDY) = 0;
	mem16(ADDER+SDX) = 0;
	mem16(ADDER+SDY) = hgt;
	mem16(ADDER+FSC) = 0x1fff;
	mem16(ADDER+SSC) = 0x1fff;
	mem16(ADDER+MDE) = 0xc0;
	mem16(ADDER+CMD) = 0x7a4;
	r9 = wid*hgt;
	r11 = ADDER+STAT;
	r10 = ADDER+IDD;
	r8 = (int)buf;
	asm("gpzl1: bbc $6,(r11),gpzl1");
	asm("      movw (r8)+,(r10)");
	asm("      sobgtr r9,gpzl1");
/*
	while(i--) {
		while(!(mem16(ADDER+STAT) & bit(6)))
			continue;
		mem16(ADDER+IDD) = *buf++;
	}
*/
}



gget(buf, x,y,wid,hgt)
	short *buf;
	int x,y,wid,hgt;
{
register int r11,r10,r9,r8;
int t;
	t = 30000;
	while(!(mem16(ADDER+STAT) & bit(3)) && t--)
		continue;
	avss_id_write(0x88, 0xffff);	/* mask 1 & mask 2 */
	avss_id_write(0x8a, 0xffff); 	/* source register */
	avss_id_write(0x8e, 0x0000);	/* background */
	avss_id_write(0x8f, 0xffff);	/* foreground */
	while(!(mem16(ADDER+STAT) & bit(4)))
		continue;
	while(!(mem16(ADDER+STAT) & bit(6)))
		continue;
	mem16(ADDER+SXO) = x;
	mem16(ADDER+SYO) = y;
	mem16(ADDER+FSDX) = wid;
	mem16(ADDER+SSDY) = hgt;
	mem16(ADDER+DXO) = 0;
	mem16(ADDER+DYO) = 0;
	mem16(ADDER+FDX) = wid;
	mem16(ADDER+FDY) = 0;
	mem16(ADDER+SDX) = 0;
	mem16(ADDER+SDY) = hgt;
	mem16(ADDER+FSC) = 0x1fff;
	mem16(ADDER+SSC) = 0x1fff;
	mem16(ADDER+MDE) = 0xc0;
	mem16(ADDER+CMD) = 0xbc4;
	r9 = (wid*hgt)/16;
	r11 = ADDER+STAT;
	r10 = ADDER+IDD;
	r8 = (int)buf;
	asm("ggl1: bbc $5,(r11),ggl1");
	asm("      movw (r10),(r8)+");
	asm("      sobgtr r9,ggl1");
/*
	while(i--) {
		while(!(mem16(ADDER+STAT) & bit(5)))
			continue;
		*buf++ = mem16(ADDER+IDD);
	}
*/
}

