/* kims_prom.s
 *   various console I/O routines for the KIMS proms
 */



/* flags and such: */

.globl _crnl
_crnl:	.long	0x1	/* 1 = enable CR/NL translations */




.globl _prom_kbd_in
_prom_kbd_in:	.word 0x0
		movl	$0x20048a75,*$0x5fe830
		ret


.globl _prom_scr_out
_prom_scr_out:	.word 0x0
		movl	$0x005fd974,*$0x5fe834
		ret


.globl _prom_ser_in
_prom_ser_in:	.word 0x0
		movl	$0x20048bb8,*$0x5fe830
		ret


.globl _prom_ser_out
_prom_ser_out:	.word 0x0
		movl	$0x20048ba1,*$0x5fe834
		ret



/*  ch = prom_getc() 
 */
.globl _prom_getc
_prom_getc: 	.word 0x2	/* save r1 */
pgetc_loop:	clrl	r0
		jsb	*$0x20040044
		blbc	r0, pgetc_loop
		movl	r1, r0
		ret


/*  ch = prom_putc(ch)
 */
.globl _prom_putc
_prom_putc: 	.word 0x4	/* save r2 */
		movl	4(ap),r2
		jsb	*$0x20040058
		movl	4(ap),r0
		ret


/* write a buffer to the console, with NL -> CR/LF translation */

.globl _prom_cwrite
_prom_cwrite:	.word 0xe	/* save r1,r2,r3 */
		movl	4(ap),r1	/* buffer address */
		movl	8(ap),r3	/* number of bytes */
cwriteloop:	clrl	r2
		movb	(r1)+,r2	/* get the next character */
		blbc	_crnl,cwriteit	/* if no translation.. */
		cmpb	r2,$10		/* is it a newline? */
		bneq	cwriteit
		movb	$13,r2		/* if so, print a CR first */
		jsb	*$0x20040058	/* put the character */
		movb	$10,r2		/* and restore the NL */
cwriteit:	jsb	*$0x20040058	/* put the character */
		sobgtr	r3,cwriteloop	/* keep looping */
		movl	8(ap),r0	/* return buffer size */
		ret
		


/* read a buffer from the console, and return
 * how many characters could actually be read (0 if no keys hit)
 * translate carriage return to newline
 */

.globl _prom_cread
_prom_cread:	.word 0x3e	/* save r1,r2,r3,r4,r5 */
		movl	4(ap),r5	/* buffer address */
		movl	8(ap),r3	/* number of bytes */
		clrl	r4		/* count of characters read */
creadloop:	movl	12(ap),r0
		jsb	*$0x20040044	/* get the character, if any */
		blbc	r0, creaddone	/* if none, end */
		blbc	_crnl,creadit	/* if no translations ... */
		cmpb	r1,$13		/* else if it's a CR */
		bneq	creadit
		movb	$10,r1		/* change it to an NL */
creadit:	movb	r1,(r5)+	/* store it */
		incl	r4		/* increase character count */
		sobgtr	r3,creadloop	/* and get some more */
creaddone:	movl	r4,r0		/* return num chars read */
		ret

