/* kims_serial.c
 *   KIMS serial I/O routines 
 *   (including keyboard decoding for channel 0,
 *    and mouse handling for channel 1)
 */


#include "types.h"
#include "serial.h"
#include "lk201.h"


#define SER_CSR	0x200a0000
#define SER_BUF 0x200a0004
#define SER_LPR	0x200a0004
#define SER_TCR	0x200a0008
#define  LLBK_2		0x800
#define  DTR_2		0x400
#define  DSRS_2		0x200
#define  RTS_2		0x100
#define SER_MSR	0x200a000c
#define SER_TDR	0x200a000c

#define INT_REQ 0x2008000f
#define INT_MSK 0x2008000c
#define INT_CLR 0x2008000f
#define	 IMASK_SR	0x80
#define	 IMASK_ST	0x40
#define	 IMASK_NP	0x20
#define	 IMASK_NS	0x10
#define	 IMASK_VF	0x08
#define	 IMASK_VS	0x04
#define  IMASK_SC	0x02
#define  IMASK_DC	0x01



#define LK201_KEY_TAB	0x20040080
#define LK201_KEY_MAP	0x20040084



extern void ser_read_isr(), ser_write_isr(), clock_isr();


int ser_ibuf[4], ser_obuf[4];
int timer = 0;


int serial_getbufsize(chan)
{
	return(buf_getsize(ser_ibuf[chan]));
}


int serial_read(chan, buf, len)
	int chan;
	uint32 buf;
	int len;
{
	return(buf_read(ser_ibuf[chan], buf, len));
}

int serial_write(chan, buf, len)
	int chan;
	uint32 buf;
	int len;
{
int n;
	n = buf_write(ser_obuf[chan], buf, len);
	mem8(SER_TCR) |= (1 << chan);		/* select xmit line */
	return(n);
}


kb_write(buf, len)
	uint32 buf;
	int len;
{
	serial_write(0, buf, len);
}


mouse_write(buf, len)
	uint32 buf;
	int len;
{
	serial_write(1, buf, len);
}



long old_ipl;

serial_write_polled(chan, buf, len)
	int chan;
	unsigned char *buf;
	int len;
{
int line;
uint32 tcr;
int t;
uint32 data;
	while(len--) {
		asm("mfpr $0x12,_old_ipl");	/* save IPL */
		asm("mtpr $5,$0x12"); 		/* set IPL to 5 */
		mem8(INT_MSK) &= ~IMASK_ST;
		tcr = 0;
		t = 30000;
		mem8(SER_TCR) |= (1 << chan);
		while(1) {
			while(((data = mem16(SER_CSR)) & bit(15)) == 0 && t--)
				continue;
			mem16(SER_CSR) &= ~bit(15);
			line = (data >> 8) & 3;
			if (line == chan)
				break;
			tcr |= (1 << line);
			mem8(SER_TCR) &= ~(1 << line);
		}
		mem8(SER_TDR) = (*buf++) & 0xff;
		while(1) {
			while(((data=mem16(SER_CSR)) & bit(15)) == 0 && t--)
				continue;
			line = (data >> 8) & 3;
			if (line == chan)
				break;
			tcr |= (1 << line);
			mem8(SER_TCR) &= ~(1 << line);
		}
		mem8(SER_TCR) &= ~(1 << chan);
		if (!tcr) {
			mem8(INT_CLR) = IMASK_ST;
			mem8(INT_MSK) |= IMASK_ST;
		} else {
			mem8(INT_MSK) |= IMASK_ST;
			mem8(SER_TCR) |= tcr;
		}
		asm("mtpr _old_ipl,$0x12");	/* restore IPL */
	}
}



/*  keycode processing routines
 *  returns 0x00 - 0x7f for normal characters
 *   and 0x80 - 0xff for extended characters
 *    0x80+1 to 0x80+20 = F1 through F20 
 */

int shift_mode = 0x00;	/* 0==normal, 1==CTRL, 2==SHIFT, 3==SHIFTLOCK */
int shift_lock = 0;	/* 0==normal, 1==SHIFTLOCK */

int keycode_process(key)
	int key;
{
register int index;
static int oldkey = 0;
	switch(key) {
		case KEY_CTRL:
			shift_mode = 1; break;
		case KEY_SHIFT:
			shift_mode = 2; break;
		case KEY_LOCK:
			if(shift_mode == 3) {
				shift_mode = 0;
				shift_lock = 0;
				kb_write("\021\204", 2);  /* x11x84 */
			} else {
				shift_mode = 3;
				shift_lock = 1;
				kb_write("\023\204", 2); /* x13x84 */
			} 
			break;
		case KEY_RELEASE:		/* keep SHIFT LOCK if enabled */
			if(!shift_lock)
				shift_mode = 0;
			else
				shift_mode = 3;
			break;
		case KEY_REPEAT:
			key = oldkey;
		default:
			oldkey = key;
			if(key<0xbc) {
				if (key >= KEY_F1 && key <= KEY_F5 && shift_mode == 1)
					key += 8;
				key += 0x40;
			} else {
				index = mem8(mem32(LK201_KEY_MAP)+(key-0xbc));
				key = mem8(mem32(LK201_KEY_TAB)
						+index*4+shift_mode);
				if(key==',' && shift_mode == 2)
					key = '<';
				if(key=='.' && shift_mode == 2)
					key = '>';
			}	
			return(key);
	}
	return(-1);
}


void ser_read_svc()
{
register int chan;
int data;
	while((data = mem16(SER_BUF)) & bit(15)) {
		if (data & 0x6000)
			continue;
		chan = (data >> 8) & 0x03;
		data &= 0xff;
		if(!chan) {
			if ((data = keycode_process(data)) < 0)
				continue;
		}
		buf_write(ser_ibuf[chan], &data, 1);
	}
}


void ser_write_svc()
{
register int stat;
register int n;
register int chan;
unsigned char data;
	while((stat = mem16(SER_CSR)) & bit(15)) {
		chan = (stat >> 8) & 0x03;
		data = 0;
		if(!(n = buf_read(ser_obuf[chan], &data, 1))) {
			mem8(SER_TCR) &= ~(1 << chan);	/* no xmit */
			break; 
		}
		mem8(SER_TDR) = data;
	}
}



serial_setline(port, rate, bits, parity)
	int port;
	int rate;
	int bits;
	int parity;
{
	mem16(SER_LPR) = 0x1000 | port | (rate << 8) | parity 
		| ((bits-5) << 3);
}


serial_break(port)
	int port;
{
int i;
	mem16(SER_TDR) = (1 << (8+port));
	msleep(1000);
	mem16(SER_TDR) = 0;
}


clock_svc()
{
	timer++;
}


msleep(msec)
	int msec;
{
int t;
int *tp;
	t = timer;
	tp = &timer;
	while(*tp < (t + (msec/10-1)+1))
		continue;
}



uint32 *scb;
uint32 istack;


serial_init()
{
#define ISTACKSIZE 4096
#define BUFSIZE 65536
int i;
uint32 buffer;
char buf[80];
	istack = smem_alloc(ISTACKSIZE)+ISTACKSIZE-4;
	buffer = smem_alloc(BUFSIZE);
	buf_init(buffer, BUFSIZE, 16);
	for(i=0; i<4; i++) {
		ser_ibuf[i] = buf_open();
		ser_obuf[i] = buf_open();
	}
	asm("mtpr _istack, $0x04"); 
	scb = (uint32*)0x5ff200;
	asm("movl  _scb,r11");
	asm("movl  $_ser_read_isr+0x01, 0x2c0(r11)");
	asm("movl  $_ser_write_isr+0x01, 0x2c4(r11)");
	asm("movl  $_clock_isr+0x01, 0x0c0(r11)");
	asm("mtpr _scb, $0x11");
	asm("mtpr $0x40, $0x18"); /* enable timer int */
	mem16(SER_CSR) = 0x0030;
	while(mem16(SER_CSR) & 0x0010)
		continue;
	mem16(SER_CSR) = 0x0020; 
	mem16(SER_TCR) = DTR_2 | RTS_2;
	serial_setline(0, BR4800, 8, PAR_NONE);
	serial_setline(1, BR4800, 8, PAR_ODD);
	serial_setline(2, BR9600, 7, PAR_EVEN);
	serial_setline(3, BR9600, 8, PAR_NONE);
	mem8(INT_CLR) = (IMASK_SR | IMASK_ST);  
	mem8(INT_MSK) |= (IMASK_SR | IMASK_ST); 
	asm("mtpr $0, $0x12");	/* lower IPL to 0 */
/*
	kb_write("\x13\x1b\x07\xb9\x23\x07\xa7\x05\x83\x11",10);	
	kb_write("\0xfd\0xd3\x13\x8f\x1b\x00\x23\x00\xe3\xa7\x8b\x11\x8f",13);	
*/
	kb_write("\375\343\023\217\033\000\043\000\343\247\213\021\217", 13);
	i = 4;
	msleep(150);
	if (serial_read(0, buf, 4) < 4)
		con_puts("*** kb init failed\r\n");
	while(i--) {
		mouse_write("\124", 1); /* 0x54 */
		msleep(600);
		if (serial_read(1, buf, 10)) {
			mouse_write("\123", 1); /* 0x53 */
			msleep(100);
			mouse_write("\122", 1); /* 0x52 */
			break;
		} else if (!i)
			con_puts("*** mouse init failed\r\n");
	}
}

getpid()
{
	return(1);
}

