/* KMON  
 *  simple recursive RPN monitor for KIMS systems
 */


#include "stdio.h"


typedef unsigned char  uchar;
typedef unsigned short ushort;
typedef unsigned long  ulong;
typedef unsigned char  boolean;



#define MAXSTACK 16384
#define MAXSYM	 512


boolean quit;
ulong *stack;
int sp;
char **symstr;
ulong *symval;
int nsym;





char *mkstring(str)
	char *str;
{
char *buf;
	buf = (char*)smem_alloc(strlen(str)+1);
	strcpy(buf, str);	
	return(buf);
}



void push(n)
	ulong n;
{
	stack[sp++] = n;
}


ulong pop()
{
	if (sp)
		return(stack[--sp]);
	else {
		printf("? stack\r\n");
		return(0);
	}
}


ulong symadr(str)
	char *str;
{
int i;
	i = 0;
	while(i < nsym) {
		if (strcmp(str, symstr[i]) == 0)
			break;
		i++;
	}
	if (i >= nsym) {
		symstr[i] = mkstring(str);
		symval[i] = 0;
		nsym++;
	}
	return((ulong)&symval[i]);
}




kmon_exec(str)
	char *str;
{
char ch;
ulong n, q;
char buf[256];
char *bp;
	while(ch = *str++) switch(ch) {
		case '+':
			n = pop();
			push(pop() + n);
			break;
		case '-':
			n = pop();
			push(pop() - n);
			break;
		case '*':
			n = pop();
			push(pop() * n);
			break;
		case '/':
			n = pop();
			push(pop() / n);
			break;
		case '%':
			n = pop();
			push(pop() % n);
			break;
		case '&':
			n = pop();
			push(pop() & n);
			break;
		case '|':
			n = pop();
			push(pop() | n);
			break;
		case '^':
			n = pop();
			push(pop() ^ n);
			break;
		case '>':
			if (*str == '=') {
				str++;
				n = pop();
				push(pop() > n);
			} else {
				n = pop();
				push(pop() >= n);
			}
			break;
		case '<':
			if (*str == '=') {
				str++;
				n = pop();
				push(pop() < n);
			} else {
				n = pop();
				push(pop() <= n);
			}
			break;
		case '"':
			bp = buf;
			while((ch = *str++) && (ch != '"'))
				*bp++ = ch;
			*bp = '\0';
			push(mkstring(buf));
			break;
		case '@':
			if (*str == 'b') {
				push(*(uchar*)pop());
				str++;
			} else if (*str == 'w') {
				push(*(ushort*)pop());
				str++;
			} else
				push(*(ulong*)pop());
			break;
		case '=':
			n = pop();
			q = pop();
			if (*str == 'b') {
				printf("*(char*)%x = %x\r\n", q, n);
				*(uchar*)q = n;
				str++;
			} else if (*str == 'w') {
				*(ushort*)q = n;
				printf("*(short*)%x = %x\r\n", q, n);
				str++;
			} else {
				*(ulong*)q = n;
				printf("*(long*)%x = %x\r\n", q, n);
			}
			break;
		case '(':
			kmon_exec((char*)(*(ulong*)pop()));
			break;
		case ')':
			break;
		case '[':
			n = pop();
			push(n);
			push(n);
			break;
		case ']':
			pop();
			break;
		case '?':
			q = pop();
			n = pop();
			if (pop())
				kmon_exec((char*)(*(ulong*)n));
			else
				kmon_exec((char*)(*(ulong*)q));
			break;
		case '.':
			n = pop();
			push(n);
			printf("%8x (%d)\r\n", n, n); 
			break;
		case ',':
			n = pop();
			q = pop();
			push(n);
			push(q);
			break;
		case '$':
			switch(*str++) {
				case 'f':
					n = pop();
					printf((char*)n, (char*)pop());
					break;
				case 'v':
					n = pop();
					avss_id_write(pop(), n);
					break;
				case 'q':
					quit++;
					return;
			}
			break;
		case 4:		/* ctrl-d */
			quit++;
			return;
		default:
			str--;
			if (ch >= '0' && ch <= '9') {
				push(strtol(str, &str, 0));
			} else {
				bp = buf;
				while((ch = *str) && ((ch >= 'a' && ch <= 'z') 
					|| (ch >= 'A' && ch <= 'Z') 
					|| (ch >= '0' && ch <= '9')
					|| (ch == '_'))) {
					*bp++ = ch;
					str++;
				}
				*bp = '\0';
				if (buf[0])
					push(symadr(buf));
				else if (ch > 32) {
					printf("? %c(%2x)%s", ch, str);
					str++;
				} else
					str++;
			}
	}
}




readline(str, len)
	char *str;
	int len;
{
int chan;
int x;
char ch;
	x = 0;
	do {
		ch = getch();
		if (ch >= 32 && x < len-1) {
			str[x++] = ch;
			putchar(ch);
		}
		if (ch == 8 || ch == 127 && x) {
			x--;
			puts("\010 \010");
		}
	} while (ch != 13 && ch != 4);
	puts("\r\n");
	if(ch == 4) {
		str[0] = 4;
		x = 1;
	}
	str[x] = '\0';
}


kmon()
{
char buf[200];
	quit = 0;
	stack = (ulong*)smem_alloc(MAXSTACK);
	sp = 0;
	symstr = (char**)smem_alloc(sizeof(char*)*MAXSYM);
	symval = (ulong*)smem_alloc(sizeof(ulong)*MAXSYM);
	nsym = 0;
	while(!quit) {
		printf("KMON> ");
		readline(buf, 200);
		kmon_exec(buf);
	}
	printf("done\r\n");
}

