/*  vsprintf.c
 *     virtual printf to a string buffer
 *     author: Frank Barrus
 */


/*  the following format characters are supported:
 *	%d	decimal
 *	%u	unsigned decimal
 *	%o	unsigned octal
 *	%x	unsigned hex
 *	%b	unsigned binary
 *	%X	capital unsigned hex
 *	%s	string
 *	%c	character
 *	%n	the number of characters written is stored
 *		in the integer pointed to by the field
 * an optional field width can be specified after the %,
 * and a leading zero implies padding with zeros
 * a negative field width will left justify the field
 */


char *base_binary= "\00201";
char *base_octal = "\01001234567";
char *base_decimal = "\0120123456789";
char *base_hex = "\0200123456789abcdef";
char *base_HEX = "\0200123456789ABCDEF";

/* sprint_int()  takes a string buffer, 
 *               a number, a base-string, and a minimum number of digits
 *		 and prints the number to the string, and returns
 *	         the number of digits
 */

int sprint_int(s, num, base, digits)
		char *s;
		long num;
		char *base;
		int digits;
{
register int q, r;
register int len;
register int x;
	x = 0;
	if (num < 0)  {
		*s++ = '-';
		num = -num;
		x = 1;
	}
	q = num / *base;
	r = num % *base;
	if (digits > 1 || q > 0)
		len = sprint_int(s, q, base, digits-1);
	else
		len = 0;
	s[len] = base[r+1];
	return(len+1+x);
} 



/* sprint_uint() does the same for unsigned integers
 */

int sprint_uint(s, num, base, digits)
		char *s;
		unsigned long num;
		char *base;
		int digits;
{
register unsigned long q, r;
register int len;
	q = num / *base;
	r = num % *base;
	if (digits > 1 || q > 0)
		len = sprint_uint(s, q, base, digits-1);
	else
		len = 0;
	s[len] = base[r+1];
	return(len+1);
} 



int sprint_2(s, num, digits)
	char *s;
	unsigned long num;
	int digits;
{
	return(sprint_uint(s, num, base_binary, digits));
}



int sprint_8(s, num, digits)
	char *s;
	unsigned long num;
	int digits;
{
	return(sprint_uint(s, num, base_octal, digits));
}



int sprint_10(s, num, digits)
	char *s;
	unsigned long num;
	int digits;
{
	return(sprint_int(s, num, base_decimal, digits));
}



int sprint_16(s, num, digits)
	char *s;
	unsigned long num;
	int digits;
{
	return(sprint_uint(s, num, base_hex, digits));
}
	



/* vsprintf()   takes a string buffer, a format string
 *  		and a pointer to a list of arguments
 *		which are written to the buffer
 *		it returns the number of characters actually written
 */


int vsprintf(s, fmt, ap)
	char *s;
	char *fmt;
	long *ap;
{
register int len;
register long arg;
register int n;
register int left;
register int size;
int i;
	len = 0;
	while(*fmt) if (*fmt == '%') {
		arg = *ap++;
		++fmt;
 		n = 0;
		left = 0;
		if (*fmt == '-')
			left = *fmt++;
		while(*fmt >= '0' && *fmt <= '9') 
			n = n * 10 + (*fmt++ - '0');
		size = 0;
		switch(*fmt++) {
	    	    case 's':
			if (!left) {
				i = n - strlen((char*)arg);
				while(i-- > 0) {
					*s++ = ' ';
					len++;
				}
			}
			i = n;
		   	while(*(char*)arg) {
				*s++ = *(char*)arg++;
				len++;
				i--;
			}
			while(left && (i-- > 0)) {
				*s++ = ' ';
				len++;
			}
			break;
		    case 'c':
			*s = (char)arg;
			size = 1;
			break;
		    case 'd':
			size = sprint_int(s, arg, base_decimal, n);
			break;
		    case 'u':
			size = sprint_uint(s, arg, base_decimal, n);
			break;
		    case 'b':
			size = sprint_uint(s, arg, base_binary, n);
			break;
		    case 'o':
			size = sprint_uint(s, arg, base_octal, n);
			break;
		    case 'x':
			size = sprint_uint(s, arg, base_hex, n);
			break;
		    case 'X':
			size = sprint_uint(s, arg, base_HEX, n);
			break;
		    case 'n':
			*(int*)arg = len;
			break;
		}	
		len += size;
		s += size;
	} else {
		*s++ = *fmt++;
		len++;
	}
	*s = '\0';
	return(len);
}



