/* vt100.c
 * simple vt100 emulator
 *
 * original by feb - Dec 1991
 *
 * modified to correctly handle "special" key characters (ala VT200)
 * 	gfk - Fri Jan 31 19:55:44 EST 1992
 */

#include "types.h"
#include "stdio.h"
#include "serial.h"
#include "lk201.h"

extern int timer;
extern char VER_MAJOR, VER_MINOR;

char VER_MAJOR =	1;
char VER_MINOR =	23;

int debug;

#define MAXSERBUF 1024	/* if buffer overflows this, send a CTRL-S */
#define MINSERBUF 132	/* once buffer is under this, send CTRL-Q */


/* configuration structure:
 *  used to store array of configuration parameters
 */ 

struct config {
	char *name;
	char val[32];
	void (*sfunc)();
};


#define nconfig 22
struct config config[nconfig];


char *get_config(name)
	char *name;
{
int i;
	i = 0;
	while(i < nconfig && strcmp(config[i].name, name))
		i++;
	if (i == nconfig)
		return((char*)0);
	else
		return(config[i].val);
}


void set_config(name, val)
	char *name;
	char *val;
{
int i;
	i = 0;
	while(i < nconfig && strcmp(config[i].name, name))
		i++;
	if (i != nconfig) {
		strcpy(config[i].val, val);
		config[i].sfunc();
	}
}



/* io_mode variables control redirection of serial lines 2 and 3, the keyboard,
 * and the console screen as follows:
 *    bit pattern:     data flow directed to:
 *	001		console screen
 *	010		serial line 2 (aux port)
 *      100		serial line 3 (console port)
 * index into io_mode array:
 *	0		keyboard input
 *	1		mouse input
 *	2		serial line 2 input (aux port)
 *	3		serial line 3 input (console port)
 */
static int io_mode[4];

static int mouse_x, mouse_y;
static int mouse_button, old_button;

static short cursor[4] = {0xffff, 0xffff, 0xffff, 0xffff};

static int keypad = 0;	/* keypad: 0==application, 1==numeric */

/* scroll-back buffer: */

char *sback_buf;
int sback_size;
int sback_pos;
int sback_vpos;
int sback_mode;	/* 0 = normal, 1 = sback w/ esc, 2 = sback w/o esc */
int hold, wait;
int scrsave, scrp1, scrp2;
int scrselect;
short *scrbuf;

main()
{
register int debug_mode; 
	stdin = 3;
	stdout = 3;
	smem_init();
	con_init();
	con_cls();
	con_printf("\033[5;7m  ShagTerm v%d.%02d  \033[0m\r\n", 
		VER_MAJOR, VER_MINOR);
	con_puts("April 5th, 1992\r\n");
	serial_init();
	asm("mtpr $0, $0x12");
	con_puts("Serial I/O initialized\r\n");
	if(debug_mode == 4)
		kmon();
	stdin = 0;
	stdout = 0;
	vt100();
}


vt100()
{
unsigned char buf[MAXSERBUF];
int len;
int i;
unsigned char ch;
	mouse_x = 0; mouse_y = 0; mouse_button = 0;
	scrsave = 0;
	scrselect = 0;
	scrbuf = (short*)smem_alloc(1536*1152/8);
	sback_mode = 0;
	sback_size = 256*1024;
	sback_pos = 0;
	sback_vpos = 0;
	sback_buf = (char*)smem_alloc(sback_size);
	sback_write("\r\n\n****** start of scrollback ******\r\n\n", 39);
	setup_init();
	hold = 0;
	wait = 0;
	while (1) {

		if(timer >= scrsave)
			screen_save(scrselect);

		/* Read the keyboard: */
		len = serial_read(0, buf, 80, 1);
		i = 0;
		while (i < len) {
			timer = 0;
			ch = buf[i++];
			if (ch >= 0x80)
				do_func_key(ch-0x40);
			else 
				vt_write(0,&ch,1);
		}
		if (hold)	/* if hold is pressed, ignore serial */
			continue;
		
		/* Read the console port: */
		len = serial_read(3, buf, 1, 1);
		if (len)
			vt_write(3,buf,len);

		/* Read the mouse: */
		len = serial_read(1, buf, 1, 1);
		if (len && mouse_process(buf,len)) {
			timer = 0;
			if (mouse_x < 0) 
				mouse_x = 0;
			if (mouse_x > 1535)
				mouse_x = 1535;
			if (mouse_y < 0)
				mouse_y = 0;
			if (mouse_y > 1151)
				mouse_y = 1151;
			if ((mouse_button & 2) && !(old_button & 2))
				cursor[0]++;
			if (mouse_button & 1)
				gput_z(cursor, mouse_x, mouse_y, 1, 1);
			avss_cursor_move(mouse_x, mouse_y);
			avss_cursor_mode(mouse_button >> 2);
			old_button = mouse_button;
		}
		/* Read the serial port: */
		if(len = serial_read(2, buf, MAXSERBUF)) {
			timer = 0;
			if (len >= MAXSERBUF - 2) {
				serial_write(2, "\023", 1);	/* ^S */
				kb_write("\023\201", 2);
				wait = 1;
			}
			sback_write(buf, len); 
			vt_write(2,buf,len); 
		}
		if (wait && len < MINSERBUF) {
			serial_write(2, "\021", 1);	/* ^Q */
			kb_write("\021\201", 2);
			wait = 0;
		}
	
	}
}



int vt_write(ichan,buf,len)
	int ichan;
	char *buf;
	int len;
{
	if (io_mode[ichan] & 1)
		con_write(buf, len);
	if (io_mode[ichan] & 2)
		serial_write(2, buf, len);
	if (io_mode[ichan] & 4)
		serial_write(3, buf, len);
	return(len);
}


int mouse_process(buf, len)
	char *buf;
	int len;
{
static int bnum = 0;
static int bx,by;
unsigned char ch;
int ok = 0;
	while(len--) {
		ch = *buf++;
		if (ch & 0x80) {
			bnum = 1;
			bx = -1;
			by = 1;
			if (ch & 0x10)
				bx = 1;
			if (ch & 0x08)
				by = -1;
			mouse_button = ch & 7;
			continue;
		}
		if (bnum == 1) {
			bx *= ch;
			bnum++;
		} else if (bnum == 2) {
			by *= ch;
			bnum = 0;
			mouse_x += bx;
			mouse_y += by;
			ok = 1;
		}
	}
	return(ok);
}


static unsigned char unblank_key;


int unblank()
{
char ch;
	unblank_key = 0;
	if(serial_read(0,&unblank_key,1) | serial_read(3,&ch,1))
		return(1);
	if(serial_read(1,&ch,1)) {
		mouse_process(&ch,1);
		return(1);
	}
	if(serial_read(2,&ch,1)) {
		vt_write(2,&ch,1);
		return(1);
	}
	return(0);
}


screen_save(mode)
	int mode;
{
	gget(scrbuf,0,0,1536,1152);
	switch(mode) {
		case 0: 
			stars(scrp2, scrp1);
			break;
		case 1:
			design(scrp1, scrp2);
			break;
		case 2:
			trek(0,0,0,1536,1152);
			break;
		case 3:
			trek(1,0,0,1536,1152);
			break;
	}
	if(unblank_key-0x40 == KEY_CTRL_F4) {
		timer = scrsave;
		if (++scrselect > 3)
			scrselect = 0;
	}	
	gput(scrbuf,0,0,1536,1152);
	timer = 0;
}




sback_write(buf, len)
	char *buf;
	int len;
{
	while(len--) {
		sback_buf[sback_pos++] = *buf++;
		if(sback_pos >= sback_size)
			sback_pos = 0;
	}
	sback_vpos = sback_pos;
}


sback_up(i)
	int i;
{
int n;
	while(i--) {
		n = con_getwidth();
		while(n-- && sback_vpos != sback_pos + 1
				&& sback_buf[sback_vpos--] != '\n')
			if (sback_vpos < 0)
				sback_vpos = sback_size-1; 
	}
}


sback_down(i)
	int i;
{
int n;
	while(i--) {
		n = con_getwidth();
		while(n-- && sback_vpos != sback_pos
				&& sback_buf[sback_vpos++] != '\n')
			if (sback_vpos >= sback_size)
				sback_vpos = 0; 
	}
}

sback_view(pos)
	int pos;
{
int i, n;
char ch;
char *bp;
char buf[132];
int wid,hgt;
	wid = con_getwidth();
	hgt = con_getheight();
	i = hgt - 1;
	while(i--) {
		n = con_getwidth();
		while(n-- && sback_buf[pos--] != '\n')
			if (pos < 0)
				pos = sback_size-1; 
			
	}
	rawcon_clear(0,0,wid-1,hgt-1);
	i = 0;
	while(i < hgt) {
		bp = buf;
		n = wid;
		while((ch = sback_buf[pos++]) != '\n' && --n) {
			if (pos >= sback_size)
				pos = 0;
			if (ch < 32) 
				if (sback_mode == 1) {
					*bp++ = '^';
					ch += 64;
					n--;
				} else {
					n++;
					if(ch==8 && bp != buf) {
						bp--;
						n++;
					}
					continue;	
				}
			*bp++ = ch;
		}
		*bp++ = '\0';
		rawcon_write(0,i,0,buf,strlen(buf));
		i++;
	}
}



do_func_key(key)
	int key;
{
int i;
int n;
char ch;
char *cp;
	switch(key) {

/** start of top-row function key definitions **/

		case KEY_F1:	/* hold screen */
			if (hold) {
				serial_write(2, "\021", 1);	/* ^Q */
				kb_write("\021\210", 2);
			} else {
				serial_write(2, "\023", 1);	/* ^S */
				kb_write("\023\210", 2);
			}
			hold = !hold;
			break;
		case KEY_F2:	/* pr_scr (redraws screen) */
			con_redraw();
			break;
		case KEY_F3:	/* go into SETUP mode */
			setup();
			break;
		case KEY_CTRL_F3:	/* enter monitor */
			kmon();
			break;
		case KEY_F4:
			timer = scrsave;
			break;
		case KEY_CTRL_F4:
			timer = scrsave;
			if (++scrselect > 3)
				scrselect = 0;
			break;
		case KEY_F5:	/* send break */
			kb_write("\023\201", 2);
			serial_break(2);
			kb_write("\021\201", 2);
			break;
		case KEY_CTRL_F5:	/* answerback */
			cp = get_config("answer");
			vt_write(0,cp,strlen(cp));
			break;
/* macro hot-keys... */

		case KEY_F6:	/* Interrupt/Halt */
			cp = get_config("macro1");
			vt_write(0,cp,strlen(cp));	
			break;
		case KEY_F7:	/* Resume */
			cp = get_config("macro2");
			vt_write(0,cp,strlen(cp));	
			break;
		case KEY_F8:	/* Cancel */
			cp = get_config("macro3");
			vt_write(0,cp,strlen(cp));	
			break;
		case KEY_F9:	/* Main Screen */
			cp = get_config("macro4");
			vt_write(0,cp,strlen(cp));	
			break;
		case KEY_F10:
			con_cls();	/* clear screen */
			break;

		case KEY_F11:
			vt_write(0,"\033",1);	/* ^[ (ESC) */
			break;
		case KEY_F12:
			vt_write(0,"\010",1);	/* ^H (BS) */
			break;
		case KEY_F13:
			vt_write(0,"\012",1);	/* ^J (LF) */
			break;
		case KEY_F14:	/* delay emulator, still handles input (testing) */
			for(i=0; i<1000000; i++);
			break;

		case KEY_F15:
			con_puts("*** Sorry, no help is available\r\n");
			break; 
		case KEY_F16:	/* DO --- reset terminal */
			asm("jmp *$0x100");
			break;

		case KEY_F17:	/* toggle keypad mode */
/*
			for(i=0; i<256; i++) 
				rawcon_write(i & 0xf, i/16, 0, &i, 1);
*/
			keypad = !keypad;
			break;
		case KEY_F18:	/* toggle Scroll-Back mode */
			if (sback_mode++ == 2) {
				sback_mode = 0;
				sback_vpos = sback_pos;
				con_redraw();
			} else {
				sback_view(sback_vpos);
			} 
			break;
		case KEY_F19:	/* scroll back */
			if(sback_mode) {
				sback_up(con_getheight());
				sback_view(sback_vpos);
			}
			break;
		case KEY_F20:	/* scroll forward */
			if(sback_mode) {
				sback_down(con_getheight());
				sback_view(sback_vpos);
			}
			break;

/** start of "arrow" keypad definitions **/

		case KEY_FIND:		/* Find */
			vt_write(0,"\033[1~",4);
			break;
		case KEY_INSERT_HERE:	/* Insert Here */
			vt_write(0,"\033[2~",4);
			break;
		case KEY_REMOVE:	/* Remove */
			vt_write(0,"\033[3~",4);
			break;
		case KEY_SELECT:	/* Select */
			vt_write(0,"\033[4~",4);
			break;
		case KEY_PREV_SCREEN:	/* Prev Screen */
			vt_write(0,"\033[5~",4);
			break;
		case KEY_NEXT_SCREEN:	/* Next Screen */
			vt_write(0,"\033[6~",4);
			break;

		case KEY_LEFT:		/* left arrow */
			vt_write(0,"\033[D",3);
			break;
		case KEY_RIGHT:		/* right arrow */
			vt_write(0,"\033[C",3);
			break;
		case KEY_UP:		/* up arrow */
			vt_write(0,"\033[A",3);
			break;
		case KEY_DOWN:		/* down arrow */
			vt_write(0,"\033[B",3);
			break;

/** start of numeric keypad definitions **/

		case KEY_KP_0:
			if (!keypad) 	/* application */
				vt_write(0,"\033Op",3);
			else		/* numeric */
				vt_write(0,"0",1);
			break;
		case KEY_KP_PERIOD:
			if (!keypad) 	/* application */
				vt_write(0,"\033On",3);
			else		/* numeric */
				vt_write(0,".",1);
			break;
		case KEY_KP_ENTER:
			if (!keypad) 	/* application */
				vt_write(0,"\033OM",3);
			else		/* numeric */
				vt_write(0,"\r",1);
			break;
		case KEY_KP_1:
			if (!keypad) 	/* application */
				vt_write(0,"\033Oq",3);
			else		/* numeric */
				vt_write(0,"1",1);
			break;
		case KEY_KP_2:
			if (!keypad) 	/* application */
				vt_write(0,"\033Or",3);
			else		/* numeric */
				vt_write(0,"2",1);
			break;
		case KEY_KP_3:
			if (!keypad) 	/* application */
				vt_write(0,"\033Os",3);
			else		/* numeric */
				vt_write(0,"3",1);
			break;
		case KEY_KP_4:
			if (!keypad) 	/* application */
				vt_write(0,"\033Ot",3);
			else		/* numeric */
				vt_write(0,"4",1);
			break;
		case KEY_KP_5:
			if (!keypad) 	/* application */
				vt_write(0,"\033Ou",3);
			else		/* numeric */
				vt_write(0,"5",1);
			break;
		case KEY_KP_6:
			if (!keypad) 	/* application */
				vt_write(0,"\033Ov",3);
			else		/* numeric */
				vt_write(0,"6",1);
			break;
		case KEY_KP_COMMA:
			if (!keypad) 	/* application */
				vt_write(0,"\033O/*",4);
			else		/* numeric */
				vt_write(0,",",1);
			break;
		case KEY_KP_7:
			if (!keypad) 	/* application */
				vt_write(0,"\033Ow",3);
			else		/* numeric */
				vt_write(0,"7",1);
			break;
		case KEY_KP_8:
			if (!keypad) 	/* application */
				vt_write(0,"\033Ox",3);
			else		/* numeric */
				vt_write(0,"8",1);
			break;
		case KEY_KP_9:
			if (!keypad) 	/* application */
				vt_write(0,"\033Oy",3);
			else		/* numeric */
				vt_write(0,"9",1);
			break;
		case KEY_KP_HYPHEN:
			if (!keypad) 	/* application */
				vt_write(0,"\033Om",3);
			else		/* numeric */
				vt_write(0,"-",1);
			break;

		case KEY_KP_PF1:
			vt_write(0,"\033OP",3);
			break;
		case KEY_KP_PF2:
			vt_write(0,"\033OQ",3);
			break;
		case KEY_KP_PF3:
			vt_write(0,"\033OR",3);
			break;
		case KEY_KP_PF4:
			vt_write(0,"\033OS",3);
			break;

#ifdef	HUNTHACK
/* start of hunt(6) special "hot-key" definitions... */

		case KEY_KP_8:
			vt_write(0,"K",1);
			for(i=0; i<500; i++)
				vt_write(0,"k",1);
			break;
		case KEY_KP_5:
			vt_write(0,"J",1);
			for(i=0; i<500; i++)
				vt_write(0,"j",1);
			break;
		case KEY_KP_4:
			vt_write(0,"H",1);
			for(i=0; i<500; i++)
				vt_write(0,"h",1);
			break;
		case KEY_KP_6:
			vt_write(0,"L",1);
			for(i=0; i<500; i++)
				vt_write(0,"l",1);
			break;

/* end of hunt(6) special "hot-key" definitions... */
#endif	HUNTHACK

		default:
			con_write("<", 1);
			ch = ((key/100)%10)+48; con_write(&ch, 1);
			ch = ((key/10)%10)+48; con_write(&ch, 1);
			ch = ((key/1)%10)+48; con_write(&ch, 1);
			con_write(">", 1);
	}
}




void set_term()
{
}


void set_comm()
{
char *str;
int baud;
int parity;
int bits;
	baud = strtol(get_config("speed"), NULL, 0);
	switch(baud) {
		case 300: baud = BR300; break;
		case 600: baud = BR600; break;
		case 1200: baud = BR1200; break;
		case 2400: baud = BR2400; break;
		case 4800: baud = BR4800; break;
		case 9600: baud = BR9600; break;
		case 19200: baud = BR19200; break;
		default: baud = BR9600; break;
	}
	str = get_config("parity");
	switch(str[0]) {
		case 'e': case 'E':
			parity = PAR_EVEN; break;
		case 'o': case 'O':
			parity = PAR_ODD; break;
		default:
			parity = PAR_NONE;
	}
	bits = strtol(get_config("bits"), NULL, 0);
	serial_setline(2,baud,bits,parity);
}


void set_size()
{
	con_setwidth(strtol(get_config("cols"), NULL, 0));
	con_setheight(strtol(get_config("rows"), NULL, 0));
}


void set_io()
{
	io_mode[0] = strtol(get_config("kbd_ctrl"), NULL, 0);
	io_mode[2] = strtol(get_config("aux_ctrl"), NULL, 0);
	io_mode[3] = strtol(get_config("cons_ctrl"), NULL, 0);
}

void set_keypad()
{
	register char *cp;

	cp = get_config("keypad");

	if (!strcmp(cp,"app")) {
		keypad = 0;	/* application keypad */
	} else if(!strcmp(cp,"num")) {
		keypad = 1;	/* numeric keypad */
	}
}


void set_debug()
{
	debug = strtol(get_config("debug"), NULL, 0);
}


void set_scrsave()
{
	scrsave = strtol(get_config("scr_del"), NULL, 0)*100;
	scrp1 = strtol(get_config("scr_p1"), NULL, 0);
	scrp2 = strtol(get_config("scr_p2"), NULL, 0);
}


void set_vol()
{
unsigned char buf[4];
	buf[0] = 0x1b;
	buf[1] = 0x87 - strtol(get_config("kvolume"), NULL, 0);
	buf[2] = 0x23;
	buf[3] = 0x87 - strtol(get_config("bvolume"), NULL, 0);
	kb_write(buf,4);
}


/*
void set_adr()
{
uint32 adr;
char buf[10];
	adr = strtol(get_config("a"), NULL, 0);
	if (adr > 0) {
		sprintf(buf, "%8x", *(uint32*)adr);
		rawcon_write(80,0,0,buf,8);
	}
}

void set_val()
{
uint32 adr, value;
	adr = strtol(get_config("a"), NULL, 0);
	value = strtol(get_config("v"), NULL, 0);
	*(uint32*)adr = value;
}
*/

void set_none()
{
}



struct config config[nconfig] = {
	{"term", "vt100", set_term},
	{"keypad", "app", set_keypad},
	{"speed", "19200", set_comm},
	{"parity", "even", set_comm},
	{"bits", "7", set_comm},
	{"cols", "128", set_size},
	{"rows", "52", set_size},
	{"answer", "<none>", set_none},
	{"macro1", "<undefined>", set_none},
	{"macro2", "<undefined>", set_none},
	{"macro3", "<undefined>", set_none},
	{"macro4", "<undefined>", set_none},
	{"kbd_ctrl", "2", set_io},
	{"aux_ctrl", "1", set_io},
	{"cons_ctrl", "2", set_io},
	{"kvolume", "7", set_vol},
	{"bvolume", "7", set_vol},
	{"debug", "0", set_debug},
	{"scr_del", "300", set_scrsave},
	{"scr_p1", "3", set_scrsave},
	{"scr_p2", "40", set_scrsave},
	{"test2", "0", set_none}
};


setup_init()
{
int i;
	for(i=0; i<nconfig; i++)
		set_config(config[i].name,config[i].val);
}


setup()
{
int hgt;
char buf[128];
int i, n;
char name[10], val[32];
char *cp, *dp;
	hgt = (nconfig+1)/2;
	for(i=0; i<128; i++)
		buf[i] = ' ';
	buf[0] = buf[127] = '|';
	for(i=46-hgt; i<52; i++)
		rawcon_write(0,i,1,buf,128);
	strcpy(&buf[54],"S E T U P   M O D E");
	buf[54+19] = ' ';
	rawcon_write(0,45-hgt,7,buf,128);
	while(1) {
		for(i=0; i<nconfig; i++) {
			for(n=0; n<42; n++)
				buf[n] = ' ';
			strcpy(buf,config[i].name);
			buf[strlen(config[i].name)] = ' ';
			strcpy(&buf[10],config[i].val);
			buf[10+strlen(config[i].val)] = ' ';
			rawcon_write(20+(i/hgt)*50,48-hgt+(i%hgt),1,buf,42);
		}
		rawcon_write(20,50,1,"Option: ", 8);
		term_input(30,50,1,buf,42);
		if (!*buf)	/* just return */
			break;
		cp = buf;
		dp = name;
		while(*cp && *cp != ' ')
			*dp++ = *cp++;
		*dp = '\0';
		while(*cp == ' ')
			cp++;
		dp = val;
		while(*cp)
			*dp++ = *cp++;
		*dp = '\0';
		set_config(name, val);
	}
	con_redraw();
}


term_input(x,y,attr,buf,size)
	int x,y,attr;
	char *buf;
	int size;
{
unsigned char ch;
int len,i;
int esc;
	esc = 0;
	for(i=0; i<size; i++)
		buf[i] = ' ';
	rawcon_write(x,y,attr,buf,size);
	len = 0;
	ch = 0;
	while(ch != 13 || esc) {
		rawcon_write(x,y,attr,buf,len);
		rawcon_write(x+len,y,7," ",1);
		while(!serial_read(0, &ch, 1, 1));
		if (ch == KEY_F3+0x40) {
			len = 0;
			break;
		}
		if (ch == '\\' && !esc)
			esc++;
		else if (ch == 127 && !esc) {
			if (len)
				len--;
			rawcon_write(x+len+1,y,attr," ",1);
		} else if (ch != 13 || esc) {
			buf[len++] = ch;
			esc = 0;
		}
	}
	rawcon_write(x,y,attr,buf,len);
	buf[len] = '\0';
}




ctrl_unknown(ch)
	char ch;
{
char *cp;
	switch(ch) {
		case 5:	cp = get_config("answer");
			vt_write(0,cp,strlen(cp));
			break;
	}
}



esc_unknown(escstr)
	char *escstr;
{
	if(escstr[1] != '[')
		escstr[2] = escstr[1];
	switch(escstr[2]) {
		case '~':
			con_puts("*** terminal download - please wait...");
			asm("movl $2,r11");
			asm("movl $0x500000,r10");
			xm_uu_download();
			/* never returns..... */
			break;
		default:
			if(debug) {
				con_puts("*** unknown: ");
				print_esc(escstr);
			}
	}
}
